library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity data32to8 is
	port (
		state					: out positive;

		clk						: in  std_logic;
		reset					: in  std_logic;
		
		word_in				: in  std_logic_vector(31 downto 0);
		word_out			: out std_logic_vector(31 downto 0);
		word_read			: in  std_logic;
		word_write		: in  std_logic;
		word_ready		: out std_logic;

		mode8					: in  std_logic;

		byte_in				: in  std_logic_vector(7 downto 0);
		byte_out  		: out std_logic_vector(7 downto 0);
		byte_read 		: out std_logic;
		byte_write		: out std_logic;
		byte_ready		: in  std_logic
	);
end entity;
	
architecture behaviour of data32to8 is
	signal save_byte1, save_byte2, save_byte3, save_byte4 : std_logic; 
	type STATETYPE is (RESET_STATE, IDLE, READ1, READ2, READ3, READ4, 
				WRITE1, WRITE2, WRITE3, WRITE4, READY_STATE);
	signal state_i, state_o: STATETYPE;
begin
	save_byte1 <= '1' when state_o = READ1 else '0';
	save_byte2 <= '1' when state_o = READ2 else '0';
	save_byte3 <= '1' when state_o = READ3 else '0';
	save_byte4 <= '1' when state_o = READ4 else '0';

	process(state_o, word_read, mode8, word_write, byte_ready, word_in) begin
		case state_o is
			when RESET_STATE =>
				word_ready <= '0';
				byte_read <= '0';
				byte_write <= '0';
				byte_out <= (others => '-');
				state_i <= IDLE;
state <= 1;
			when IDLE =>
				word_ready <= '0';
				byte_read <= '0';
				byte_write <= '0';
				byte_out <= (others => '-');
				if (word_read = '1') then
					if (mode8 = '1') then
						state_i <= READ4;
					else
						state_i <= READ1;
					end if;
				elsif (word_write = '1') then
					if (mode8 = '1') then
						state_i <= WRITE4;
					else
						state_i <= WRITE1;
					end if;
				else
					state_i <= IDLE;
				end if;
state <= 2;
			when READ1 =>
				word_ready <= '0';
				byte_read <= '1';
				byte_write <= '0';
				byte_out <= (others => '-');
				if (byte_ready = '1') then
					state_i <= READ2;
				else
					state_i <= READ1;
				end if;
state <= 3;
			when READ2 =>
				word_ready <= '0';
				byte_read <= '1';
				byte_write <= '0';
				byte_out <= (others => '-');
				if (byte_ready = '1') then
					state_i <= READ3;
				else
					state_i <= READ2;
				end if;
state <= 4;
			when READ3 =>
				word_ready <= '0';
				byte_read <= '1';
				byte_write <= '0';
				byte_out <= (others => '-');
				if (byte_ready = '1') then
					state_i <= READ4;
				else
					state_i <= READ3;
				end if;
state <= 5;
			when READ4 =>
				word_ready <= '0';
				byte_read <= '1';
				byte_write <= '0';
				byte_out <= (others => '-');
				if (byte_ready = '1') then
					state_i <= READY_STATE;
				else
					state_i <= READ4;
				end if;
state <= 6;
			when WRITE1 =>
				word_ready <= '0';
				byte_read <= '0';
				byte_write <= '1';
				byte_out <= word_in(31 downto 24);
				if (byte_ready = '1') then
					state_i <= WRITE2;
				else
					state_i <= WRITE1;
				end if;
state <= 7;
			when WRITE2 =>
				word_ready <= '0';
				byte_read <= '0';
				byte_write <= '1';
				byte_out <= word_in(23 downto 16);
				if (byte_ready = '1') then
					state_i <= WRITE3;
				else
					state_i <= WRITE2;
				end if;
state <= 8;
			when WRITE3 =>
				word_ready <= '0';
				byte_read <= '0';
				byte_write <= '1';
				byte_out <= word_in(15 downto 8);
				if (byte_ready = '1') then
					state_i <= WRITE4;
				else
					state_i <= WRITE3;
				end if;
state <= 9;
			when WRITE4 =>
				word_ready <= '0';
				byte_read <= '0';
				byte_write <= '1';
				byte_out <= word_in(7 downto 0);
				if (byte_ready = '1') then
					state_i <= READY_STATE;
				else
					state_i <= WRITE4;
				end if;
state <= 10;
			when READY_STATE =>
				word_ready <= '1';
				byte_read <= '0';
				byte_write <= '0';
				byte_out <= (others => '-');
				state_i <= IDLE;
state <= 11;
		end case;	
	end process;

	process(clk, reset, state_i) begin
		if (clk'event and clk = '1') then
			if (reset = '1') then
				state_o <= RESET_STATE;
			else
				state_o <= state_i;
			end if;
		end if;
	end process;

	process(clk, reset, save_byte1, save_byte2, save_byte3, save_byte4) begin
		if (clk'event and clk = '1') then
			if (reset = '1') then
				word_out <= (others => '0');
			else
				if (save_byte1 = '1') then
					word_out(31 downto 24) <= byte_in;
				end if;
				if (save_byte2 = '1') then
					word_out(23 downto 16) <= byte_in;
				end if;
				if (save_byte3 = '1') then
					word_out(15 downto 8) <= byte_in;
				end if;
				if (save_byte4 = '1') then
					word_out(7 downto 0) <= byte_in;
				end if;
			end if;
		end if;
	end process;
end architecture;
